@@ -1,6 +1,7 @@
+import importlib.util
 from pathlib import Path
 
-VERSION = "6.0.9"
+VERSION = "6.0.10"
 USER = "knot-resolver"
 GROUP = "knot-resolver"
 
@@ -17,3 +18,11 @@ API_SOCK_FILE = RUN_DIR / "kres-api.sock"
 # executables paths
 KRESD_EXECUTABLE = SBIN_DIR / "kresd"
 KRES_CACHE_GC_EXECUTABLE = SBIN_DIR / "kres-cache-gc"
+
+WATCHDOG_LIB = False
+if importlib.util.find_spec("watchdog"):
+    WATCHDOG_LIB = True
+
+PROMETHEUS_LIB = False
+if importlib.util.find_spec("prometheus_client"):
+    PROMETHEUS_LIB = True
